@@ -1,22 +1,20 @@
 '''Script containing the main dataset class'''
 
 from pathlib import Path
-from typing import Union, Dict, Tuple, List
+from typing import Union, Dict, Tuple, List, Optional
 import pandas as pd
+from pandas.errors import PerformanceWarning
 import numpy as np
 import logging
 import requests
-import zipfile
-import io
-import shutil
 from collections import defaultdict
 import re
 import os
 import multiprocessing as mp
 from tqdm.auto import tqdm
 import warnings
-import csv
 import json
+from functools import partial
 
 from .twitter import Twitter
 from .article import process_article_url
@@ -60,21 +58,18 @@ class MuminDataset:
             Whether to include hashtags in the dataset. Defaults to True.
         include_mentions (bool, optional):
             Whether to include mentions in the dataset. Defaults to True.
-        include_places (bool, optional):
-            Whether to include places in the dataset. Defaults to True.
-        include_polls (bool, optional):
-            Whether to include polls in the dataset. Defaults to True.
         text_embedding_model_id (str, optional):
             The HuggingFace Hub model ID to use when embedding texts. Defaults
-            to 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2'.
+            to 'xlm-roberta-base'.
         image_embedding_model_id (str, optional):
             The HuggingFace Hub model ID to use when embedding images. Defaults
-            to 'facebook/deit-base-distilled-patch16-224'.
-        dataset_dir (str or pathlib Path, optional):
-            The path to the folder where the dataset should be stored. Defaults
-            to './mumin'.
+            to 'google/vit-base-patch16-224-in21k'.
+        dataset_path (str, pathlib Path or None, optional):
+            The path to the folder where the dataset should be stored. If None
+            then the dataset will be stored at './mumin-<size>.hdf'. Defaults
+            to None.
         verbose (bool, optional):
-            Whether extra information should be outputted. Defaults to False.
+            Whether extra information should be outputted. Defaults to True.
 
     Attributes:
         include_replies (bool): Whether to include replies in the dataset.
@@ -82,10 +77,8 @@ class MuminDataset:
         include_images (bool): Whether to include images in the dataset.
         include_hashtags (bool): Whether to include hashtags in the dataset.
         include_mentions (bool): Whether to include mentions in the dataset.
-        include_places (bool): Whether to include places in the dataset.
-        include_polls (bool): Whether to include polls in the dataset.
         size (str): The size of the dataset.
-        dataset_dir (pathlib Path): The dataset directory.
+        dataset_path (pathlib Path): The dataset file.
         text_embedding_model_id (str): The model ID used for embedding text.
         image_embedding_model_id (str): The model ID used for embedding images.
         nodes (dict): The nodes of the dataset.
@@ -97,29 +90,24 @@ class MuminDataset:
               Network Posts_ (2021)
     '''
     download_url: str = ('https://github.com/CLARITI-REPHRAIN/mumin-build/raw/'
-                         '25fab15b7f27e99fca34637891f18c9927609473/'
-                         'data/mumin.zip')
+                         'c5276bab7e9421daaa4e6acb7d7ee48cd966b329/'
+                         'data/mumin.hdf')
     _node_dump: List[str] = ['claim', 'tweet', 'user', 'image', 'article',
-                             'place', 'hashtag', 'poll', 'reply']
+                             'hashtag', 'reply']
     _rel_dump: List[Tuple[str, str, str]] = [
         ('tweet', 'discusses', 'claim'),
         ('tweet', 'mentions', 'user'),
-        ('tweet', 'located_in', 'place'),
         ('tweet', 'has_image', 'image'),
         ('tweet', 'has_hashtag', 'hashtag'),
         ('tweet', 'has_article', 'article'),
-        ('tweet', 'has_poll', 'poll'),
         ('reply', 'reply_to', 'tweet'),
-        ('reply', 'reply_to', 'reply'),
         ('reply', 'quote_of', 'tweet'),
         ('user', 'posted', 'tweet'),
         ('user', 'posted', 'reply'),
         ('user', 'mentions', 'user'),
-        ('user', 'has_pinned', 'tweet'),
         ('user', 'has_hashtag', 'hashtag'),
         ('user', 'has_profile_picture', 'image'),
         ('user', 'retweeted', 'tweet'),
-        ('user', 'liked', 'tweet'),
         ('user', 'follows', 'user'),
         ('article', 'has_top_image', 'image'),
     ]
@@ -132,15 +120,11 @@ class MuminDataset:
                  include_images: bool = True,
                  include_hashtags: bool = True,
                  include_mentions: bool = True,
-                 include_places: bool = True,
-                 include_polls: bool = True,
-                 text_embedding_model_id: str = ('sentence-transformers/'
-                                                 'paraphrase-multilingual-'
-                                                 'MiniLM-L12-v2'),
-                 image_embedding_model_id: str = ('facebook/deit-base-'
-                                                  'distilled-patch16-224'),
-                 dataset_dir: Union[str, Path] = './mumin',
-                 verbose: bool = False):
+                 text_embedding_model_id: str = 'xlm-roberta-base',
+                 image_embedding_model_id: str = ('google/vit-base-patch16-'
+                                                  '224-in21k'),
+                 dataset_path: Optional[Union[str, Path]] = None,
+                 verbose: bool = True):
         self.compiled = False
         self._twitter = Twitter(twitter_bearer_token=twitter_bearer_token)
         self.size = size
@@ -149,15 +133,16 @@ class MuminDataset:
         self.include_images = include_images
         self.include_hashtags = include_hashtags
         self.include_mentions = include_mentions
-        self.include_places = include_places
-        self.include_polls = include_polls
         self.text_embedding_model_id = text_embedding_model_id
         self.image_embedding_model_id = image_embedding_model_id
-        self.dataset_dir = Path(dataset_dir)
         self.verbose = verbose
         self.nodes: Dict[str, pd.DataFrame] = dict()
         self.rels: Dict[Tuple[str, str, str], pd.DataFrame] = dict()
 
+        if dataset_path is None:
+            dataset_path = f'./mumin-{size}.hdf'
+        self.dataset_path = Path(dataset_path)
+
         # Set up logging verbosity
         if self.verbose:
             logger.setLevel(logging.INFO)
@@ -165,7 +150,7 @@ class MuminDataset:
             logger.setLevel(logging.WARNING)
 
     def __repr__(self) -> str:
-        '''A string representation of the dataaset.
+        '''A string representation of the dataset.
 
         Returns:
             str: The representation of the dataset.
@@ -196,23 +181,33 @@ class MuminDataset:
 
         # Only compile the dataset if it has not already been compiled
         if 'text' not in self.nodes['tweet'].columns:
+
+            # Shrink dataset to the correct size
             self._shrink_dataset()
+
+            # Rehydrate the tweets
             self._rehydrate(node_type='tweet')
             self._rehydrate(node_type='reply')
+
+            # Update the IDs of the data that was there pre-hydration
             self._update_precomputed_ids()
+
+            # Extract data from the rehydrated tweets
             self._extract_nodes()
             self._extract_relations()
             self._extract_articles()
             self._extract_images()
+
+            # Filter the data
             self._filter_node_features()
             self._filter_relations()
 
-        # Clean up dataset
+        # Remove unnecessary bits
         self._remove_auxilliaries()
         self._remove_islands()
 
-        # Store dataset
-        self._dump_to_csv()
+        # Save dataset
+        self._dump_dataset()
 
         # Mark dataset as compiled
         self.compiled = True
@@ -220,34 +215,40 @@ class MuminDataset:
         return self
 
     def _download(self, overwrite: bool = False):
-        '''Downloads and unzips the dataset.
+        '''Downloads the dataset.
 
         Args:
             overwrite (bool, optional):
                 Whether the dataset directory should be overwritten, in case it
                 already exists. Defaults to False.
         '''
-        if (not self.dataset_dir.exists() or
-                (self.dataset_dir.exists() and overwrite)):
+        if (not self.dataset_path.exists() or
+                (self.dataset_path.exists() and overwrite)):
 
             logger.info('Downloading dataset')
 
             # Remove existing directory if we are overwriting
-            if self.dataset_dir.exists() and overwrite:
-                shutil.rmtree(self.dataset_dir)
+            if self.dataset_path.exists() and overwrite:
+                self.dataset_path.unlink()
 
-            response = requests.get(self.download_url)
+            # Set up download stream of dataset
+            with requests.get(self.download_url, stream=True) as response:
 
-            # If the response was unsuccessful then raise an error
-            if response.status_code != 200:
-                msg = f'[{response.status_code}] {response.content}'
-                raise RuntimeError(msg)
+                # If the response was unsuccessful then raise an error
+                if response.status_code != 200:
+                    msg = f'[{response.status_code}] {response.content}'
+                    raise RuntimeError(msg)
 
-            # Otherwise unzip the in-memory zip file to `self.dataset_dir`
-            else:
-                zipped = response.content
-                with zipfile.ZipFile(io.BytesIO(zipped)) as zip_file:
-                    zip_file.extractall(self.dataset_dir)
+                # Download dataset with progress bar
+                total = int(response.headers['Content-Length'])
+                with tqdm(total=total,
+                          unit='iB',
+                          unit_scale=True,
+                          desc='Downloading MuMiN') as pbar:
+                    with Path(self.dataset_path).open('wb') as f:
+                        for data in response.iter_content(1024):
+                            pbar.update(len(data))
+                            f.write(data)
 
         return self
 
@@ -259,49 +260,39 @@ class MuminDataset:
                 If the dataset has not been downloaded yet.
         '''
         # Raise error if the dataset has not been downloaded yet
-        if not self.dataset_dir.exists():
+        if not self.dataset_path.exists():
             raise RuntimeError('Dataset has not been downloaded yet!')
 
         logger.info('Loading dataset')
 
-        # Set up Pandas `read_csv` parameters
-        csv_params = dict(engine='python', on_bad_lines='skip')
-
         # Reset `nodes` and `relations` to ensure a fresh start
         self.nodes = dict()
         self.rels = dict()
 
-        # Loop over the files in the dataset directory
-        csv_paths = [path for path in self.dataset_dir.iterdir()
-                     if str(path)[-4:] == '.csv']
-        for path in csv_paths:
-            fname = path.stem
+        with pd.HDFStore(self.dataset_path) as hdf:
+            ntypes = [name.replace('/', '')
+                      for name in hdf.keys() if '_' not in name]
+            etypes = [name.replace('/', '')
+                      for name in hdf.keys() if '_' in name]
 
-            # Node case: no underscores in file name
-            if len(fname.split('_')) == 1:
-                self.nodes[fname] = pd.read_csv(path, **csv_params)
+            for ntype in ntypes:
+                self.nodes[ntype] = hdf[ntype].copy()
 
-            # Relation case: exactly two underscores in file name
-            elif len(fname.split('_')) > 2:
-                splits = fname.split('_')
+            for etype in etypes:
+                splits = etype.split('_')
                 src = splits[0]
                 tgt = splits[-1]
                 rel = '_'.join(splits[1:-1])
-                self.rels[(src, rel, tgt)] = pd.read_csv(path, **csv_params)
-
-            # Otherwise raise error
-            else:
-                raise RuntimeError(f'Could not recognise {fname} as a node '
-                                   f'or relation.')
+                self.rels[(src, rel, tgt)] = hdf[etype].copy()
 
         # Ensure that claims are present in the dataset
         if 'claim' not in self.nodes.keys():
-            raise RuntimeError('No claims are present in the zipfile!')
+            raise RuntimeError('No claims are present in the file!')
 
         # Ensure that tweets are present in the dataset, and also that the
         # tweet IDs are unique
         if 'tweet' not in self.nodes.keys():
-            raise RuntimeError('No tweets are present in the zipfile!')
+            raise RuntimeError('No tweets are present in the file!')
         else:
             tweet_df = self.nodes['tweet']
             duplicated = (tweet_df[tweet_df.tweet_id.duplicated()].tweet_id
@@ -503,28 +494,6 @@ class MuminDataset:
                                     .reset_index(drop=True))
                 self.nodes['image'] = image_df
 
-            # Extract and store polls
-            if self.include_polls and len(tweet_dfs['polls']):
-                if 'poll' in self.nodes.keys():
-                    poll_df = (self.nodes['poll']
-                                   .append(tweet_dfs['polls'])
-                                   .drop_duplicates(subset='poll_id')
-                                   .reset_index(drop=True))
-                else:
-                    poll_df = tweet_dfs['polls']
-                self.nodes['poll'] = poll_df
-
-            # Extract and store places
-            if self.include_places and len(tweet_dfs['places']):
-                if 'place' in self.nodes.keys():
-                    place_df = (self.nodes['place']
-                                    .append(tweet_dfs['places'])
-                                    .drop_duplicates(subset='place_id')
-                                    .reset_index(drop=True))
-                else:
-                    place_df = tweet_dfs['places']
-                self.nodes['place'] = place_df
-
             return self
 
     def _update_precomputed_ids(self):
@@ -820,34 +789,6 @@ class MuminDataset:
                                .reset_index(drop=True))
             self.nodes['url'] = node_df
 
-        # Add place features
-        if self.include_places and 'place' in self.nodes.keys():
-
-            def get_lat(bbox: list) -> float:
-                return (bbox[1] + bbox[3]) / 2
-
-            def get_lng(bbox: list) -> float:
-                return (bbox[0] + bbox[2]) / 2
-
-            place_df = self.nodes['place']
-            place_df['lat'] = place_df['geo.bbox'].map(get_lat)
-            place_df['lng'] = place_df['geo.bbox'].map(get_lng)
-            self.nodes['place'] = place_df
-
-        # Add poll features
-        if self.include_polls and 'poll' in self.nodes.keys():
-
-            def get_labels(options: List[dict]) -> List[str]:
-                return [dct['label'] for dct in options]
-
-            def get_votes(options: List[dict]) -> List[int]:
-                return [dct['votes'] for dct in options]
-
-            poll_df = self.nodes['poll']
-            poll_df['labels'] = poll_df.options.map(get_labels)
-            poll_df['votes'] = poll_df.options.map(get_votes)
-            self.nodes['poll'] = poll_df
-
         return self
 
     def _extract_relations(self):
@@ -935,60 +876,6 @@ class MuminDataset:
             rel_df = pd.DataFrame(data_dict)
             self.rels[('user', 'mentions', 'user')] = rel_df
 
-        # (:User)-[:HAS_PINNED]->(:Tweet)
-        pinned_exist = 'pinned_tweet_id' in self.nodes['user'].columns
-        if pinned_exist:
-            merged = (self.nodes['user'][['pinned_tweet_id']]
-                          .dropna()
-                          .reset_index()
-                          .rename(columns=dict(index='user_idx'))
-                          .astype({'pinned_tweet_id': int})
-                          .merge(self.nodes['tweet'][['tweet_id']]
-                                     .reset_index()
-                                     .rename(columns=dict(index='tweet_idx')),
-                                 left_on='pinned_tweet_id',
-                                 right_on='tweet_id'))
-            data_dict = dict(src=merged.user_idx.tolist(),
-                             tgt=merged.tweet_idx.tolist())
-            rel_df = pd.DataFrame(data_dict)
-            self.rels[('user', 'has_pinned', 'tweet')] = rel_df
-
-        # (:Tweet)-[:LOCATED_IN]->(:Place)
-        places_exist = 'geo.place_id' in self.nodes['tweet'].columns
-        if self.include_places and places_exist:
-            merged = (self.nodes['tweet'][['geo.place_id']]
-                          .dropna()
-                          .reset_index()
-                          .rename(columns=dict(index='tweet_idx'))
-                          .merge(self.nodes['place'][['place_id']]
-                                     .reset_index()
-                                     .rename(columns=dict(index='place_idx')),
-                                 left_on='geo.place_id',
-                                 right_on='place_id'))
-            data_dict = dict(src=merged.tweet_idx.tolist(),
-                             tgt=merged.place_idx.tolist())
-            rel_df = pd.DataFrame(data_dict)
-            self.rels[('tweet', 'located_in', 'place')] = rel_df
-
-        # (:Tweet)-[:HAS_POLL]->(:Poll)
-        polls_exist = 'attachments.poll_ids' in self.nodes['tweet'].columns
-        if self.include_polls and polls_exist:
-            merged = (self.nodes['tweet'][['attachments.poll_ids']]
-                          .dropna()
-                          .reset_index()
-                          .rename(columns=dict(index='tweet_idx'))
-                          .explode('attachments.poll_ids')
-                          .astype({'attachments.poll_ids': int})
-                          .merge(self.nodes['poll'][['poll_id']]
-                                     .reset_index()
-                                     .rename(columns=dict(index='poll_idx')),
-                                 left_on='attachments.poll_ids',
-                                 right_on='poll_id'))
-            data_dict = dict(src=merged.tweet_idx.tolist(),
-                             tgt=merged.poll_idx.tolist())
-            rel_df = pd.DataFrame(data_dict)
-            self.rels[('tweet', 'has_poll', 'poll')] = rel_df
-
         # (:Tweet)-[:HAS_IMAGE]->(:Image)
         images_exist = 'attachments.media_keys' in self.nodes['tweet'].columns
         if self.include_images and images_exist:
@@ -1383,26 +1270,32 @@ class MuminDataset:
             self._embed_claims()
 
         # Dump the nodes with all the embeddings
-        self._dump_to_csv()
+        self._dump_dataset()
 
         return self
 
+    @staticmethod
+    def _embed_text(text: str, tokenizer, model):
+        '''Extract a text embedding'''
+        import torch
+        with torch.no_grad():
+            inputs = tokenizer(text, truncation=True, return_tensors='pt')
+            result = model(**inputs)
+            return result.pooler_output[0].numpy()
+
     def _embed_tweets(self):
         '''Embeds all the tweets in the dataset'''
-        import transformers
+        from transformers import AutoModel, AutoTokenizer
 
         logger.info('Embedding tweets')
 
         # Load text embedding model
         model_id = self.text_embedding_model_id
-        pipeline = transformers.pipeline(task='feature-extraction',
-                                         model=model_id,
-                                         tokenizer=model_id)
+        tokenizer = AutoTokenizer.from_pretrained(model_id)
+        model = AutoModel.from_pretrained(model_id)
 
         # Define embedding function
-        def embed(text: str):
-            '''Extract a text embedding'''
-            return np.asarray(pipeline(text, truncation=True))[0, 0, :]
+        embed = partial(self._embed_text, tokenizer=tokenizer, model=model)
 
         # Embed tweet text using the pretrained transformer
         text_embs = self.nodes['tweet'].text.progress_apply(embed)
@@ -1418,20 +1311,17 @@ class MuminDataset:
 
     def _embed_replies(self):
         '''Embeds all the replies in the dataset'''
-        import transformers
+        from transformers import AutoModel, AutoTokenizer
 
         logger.info('Embedding replies')
 
         # Load text embedding model
         model_id = self.text_embedding_model_id
-        pipeline = transformers.pipeline(task='feature-extraction',
-                                         model=model_id,
-                                         tokenizer=model_id)
+        tokenizer = AutoTokenizer.from_pretrained(model_id)
+        model = AutoModel.from_pretrained(model_id)
 
         # Define embedding function
-        def embed(text: str):
-            '''Extract a text embedding'''
-            return np.asarray(pipeline(text, truncation=True))[0, 0, :]
+        embed = partial(self._embed_text, tokenizer=tokenizer, model=model)
 
         # Embed tweet text using the pretrained transformer
         text_embs = self.nodes['reply'].text.progress_apply(embed)
@@ -1447,28 +1337,25 @@ class MuminDataset:
 
     def _embed_users(self):
         '''Embeds all the users in the dataset'''
-        import transformers
+        from transformers import AutoModel, AutoTokenizer
 
         logger.info('Embedding users')
 
         # Load text embedding model
         model_id = self.text_embedding_model_id
-        pipeline = transformers.pipeline(task='feature-extraction',
-                                         model=model_id,
-                                         tokenizer=model_id)
+        tokenizer = AutoTokenizer.from_pretrained(model_id)
+        model = AutoModel.from_pretrained(model_id)
 
         # Define embedding function
-        def embed(text: Union[float, str]):
+        def embed(text: str):
             '''Extract a text embedding'''
             if text != text:
-                return None
+                return np.zeros(model.config.hidden_size)
             else:
-                return np.asarray(pipeline(text, truncation=True))[0, 0, :]
+                return self._embed_text(text, tokenizer=tokenizer, model=model)
 
         # Embed user description using the pretrained transformer
         desc_embs = self.nodes['user'].description.progress_apply(embed)
-        emb_dim = desc_embs.dropna().iloc[0].shape[-1]
-        desc_embs[desc_embs.isna()] = np.zeros((len(desc_embs), emb_dim))
         self.nodes['user']['description_emb'] = desc_embs
 
         return self
@@ -1476,26 +1363,24 @@ class MuminDataset:
     def _embed_articles(self):
         '''Embeds all the tweets in the dataset'''
         if self.include_articles:
-            import transformers
+            from transformers import AutoModel, AutoTokenizer
 
             logger.info('Embedding articles')
 
             # Load text embedding model
             model_id = self.text_embedding_model_id
-            pipeline = transformers.pipeline(task='feature-extraction',
-                                             model=model_id,
-                                             tokenizer=model_id)
+            tokenizer = AutoTokenizer.from_pretrained(model_id)
+            model = AutoModel.from_pretrained(model_id)
 
             # Define embedding function
             def embed(text: Union[str, List[str]]):
                 '''Extract a text embedding'''
+                params = dict(tokenizer=tokenizer, model=model)
                 if isinstance(text, str):
-                    return np.asarray(pipeline(text, truncation=True))[0, 0, :]
+                    return self._embed_text(text, **params)
                 else:
-                    arrays = [np.asarray(pipeline(doc,
-                                                  truncation=True))[0, 0, :]
-                              for doc in text]
-                    return np.mean(arrays)
+                    return np.mean([self._embed_text(doc, **params)
+                                    for doc in text])
 
             def split_content(doc: str) -> List[str]:
                 '''Split up a string into smaller chunks'''
@@ -1535,6 +1420,7 @@ class MuminDataset:
             def embed(image):
                 '''Extract the last hiden state of image model'''
                 with torch.no_grad():
+
                     # Ensure that the input has shape (C, H, W)
                     image = np.transpose(image, (2, 0, 1))
 
@@ -1542,16 +1428,13 @@ class MuminDataset:
                     inputs = feature_extractor(images=image,
                                                return_tensors='pt')
 
-                    # Get the embedding, being the last hidden state of the
-                    # model (we return the first sequence element, as this
-                    # corresponds to the [HEAD] tag)
-                    outputs = model(**inputs,
-                                    output_hidden_states=True,
-                                    output_attentions=True)
-                    torch_embedding = outputs.hidden_states[-1][0, 0, :]
+                    # Get the embedding
+                    outputs = model(**inputs, output_hidden_states=True)
+                    penultimate_embedding = outputs.hidden_states[-1]
+                    cls_embedding = penultimate_embedding[:, 0, :]
 
                     # Convert to NumPy and return
-                    return torch_embedding.numpy()
+                    return cls_embedding.numpy()
 
             # Embed pixels using the pretrained transformer
             with warnings.catch_warnings():
@@ -1807,25 +1690,23 @@ class MuminDataset:
 
         return self
 
-    def _dump_to_csv(self):
-        '''Dumps the dataset to CSV files'''
-        logger.info('Dumping to CSV')
+    def _dump_dataset(self):
+        '''Dumps the dataset to hdf files'''
+        logger.info('Dumping dataset')
+        hdf_path = self.dataset_path
 
-        # Dump the nodes
-        for node_type in self._node_dump:
-            if node_type in self.nodes.keys():
-                path = self.dataset_dir / f'{node_type}.csv'
-                self.nodes[node_type].to_csv(path,
-                                             index=False,
-                                             quoting=csv.QUOTE_NONNUMERIC)
+        with warnings.catch_warnings():
+            warnings.simplefilter('ignore', PerformanceWarning)
 
-        # Dump the relations
-        for rel_type in self._rel_dump:
-            if rel_type in self.rels.keys():
-                path = self.dataset_dir / f'{"_".join(rel_type)}.csv'
-                self.rels[rel_type].to_csv(path,
-                                           index=False,
-                                           quoting=csv.QUOTE_NONNUMERIC)
+            # Dump the nodes
+            for node_type in self._node_dump:
+                if node_type in self.nodes.keys():
+                    self.nodes[node_type].to_hdf(hdf_path, node_type)
+
+            # Dump the relations
+            for rel_type in self._rel_dump:
+                if rel_type in self.rels.keys():
+                    self.rels[rel_type].to_hdf(hdf_path, '_'.join(rel_type))
 
         return self
 
