@@ -46,6 +46,7 @@ def build_dgl_dataset(nodes: Dict[str, pd.DataFrame],
     # Set up the graph as a DGL graph
     graph_data = dict()
     for canonical_etype, rel_arr in relations.items():
+        src, rel, tgt = canonical_etype
         rel_arr = (relations[canonical_etype][['src', 'tgt']].drop_duplicates()
                                                              .to_numpy())
         if rel_arr.size:
@@ -53,6 +54,10 @@ def build_dgl_dataset(nodes: Dict[str, pd.DataFrame],
             tgt_tensor = torch.from_numpy(rel_arr[:, 1]).int()
             graph_data[canonical_etype] = (src_tensor, tgt_tensor)
 
+            # Adding inverse relations as well, to ensure that graph is
+            # bidirected
+            graph_data[(tgt, f'{rel}_inv', src)] = (tgt_tensor, src_tensor)
+
     dgl_graph = dgl.heterograph(graph_data)
 
     def emb_to_tensor(df: pd.DataFrame, col_name: str):
@@ -152,7 +157,7 @@ def build_dgl_dataset(nodes: Dict[str, pd.DataFrame],
 
     # Add labels
     def numericalise_labels(label: str) -> int:
-        numericalise = dict(misinformation=0, factual=1, other=2)
+        numericalise = dict(misinformation=0, factual=1)
         return numericalise[label]
     claim_labels = nodes['claim'][['label']].applymap(numericalise_labels)
     discusses = relations[('tweet', 'discusses', 'claim')]
@@ -179,11 +184,10 @@ def build_dgl_dataset(nodes: Dict[str, pd.DataFrame],
                                    right_on='src',
                                    how='left'))
     for col_name in mask_names:
-        claim_tensor = torch.from_numpy(nodes['claim'][col_name].astype(int)
-                                                                .to_numpy())
-        tweet_tensor = torch.from_numpy(merged[col_name].astype(int)
-                                                        .to_numpy())
+        claim_tensor = torch.from_numpy(nodes['claim'][col_name].to_numpy())
+        tweet_tensor = torch.from_numpy(merged[col_name].to_numpy())
         dgl_graph.nodes['claim'].data[col_name] = claim_tensor
         dgl_graph.nodes['tweet'].data[col_name] = tweet_tensor
 
+    # Convert graph to bidirected graph and return it
     return dgl_graph
